/*
 * COPYRIGHT (c) 2010 The Members of the Deacon Project <http://deacon.daverea.com/core-team/>
 * 
 * This file is part of Deacon
 * 
 * Deacon is free software: you can redistribute it and/or modify it under the terms of the GNU 
 * Lesser General Public License as published by the Free Software Foundation, either version 3 
 * of the License, or (at your option) any later version.
 * 
 * Deacon is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License and the GNU Lesser General 
 * Public License along with Deacon.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

package org.deacon;

import java.io.IOException;
import java.net.UnknownHostException;
import java.util.Set;

import android.content.BroadcastReceiver;
import android.content.Context;
import android.content.Intent;
import android.content.IntentFilter;
import android.net.NetworkInfo;
import android.os.Bundle;
import android.os.Handler;
import android.os.Message;
import android.util.Log;

/**
 * <p><strong>Deacon</strong> is the main client class for receiving Meteor push notifications on Android devices.
 * <p>Deacon is a thin wrapper around DeaconService which isolates Android-specific code to maintain standalone testability of the DeaconService class.
 * Instantiate a Deacon object within an Android app to create a new push notifications client.</p>
 */
public class Deacon extends DeaconService {

	private Context parent = null;
	
	/**
	 * Message handler used to pass received server strings from the Deacon thread (in DeaconService) to the parse() method.
	 * This is necessary in order to provide inter-thread communications within the Android environment.
	 */
	private Handler handler = new Handler() {
		@Override
		public void handleMessage(final Message msg) {
			Deacon.super.parse((String)msg.obj); // Call superclass parse() method from within inner class
		}
	};
	
	/**
	 * BroadcastReceiver for receiving CONNECTIVITY_ACTIONs
	 * Used to detect changes in network state instead of polling....
	 * http://stackoverflow.com/questions/2294971/intent-action-for-network-events-in-android-sdk
	 */
	private BroadcastReceiver bcr = new BroadcastReceiver() {
		
		@Override
		public void onReceive(Context context, Intent intent) {
			Log.d("Deacon","Intent Received");
			if (intent.getAction().equals(android.net.ConnectivityManager.CONNECTIVITY_ACTION)) {
				Log.d(Deacon.class.getSimpleName(), "action: " + intent.getAction());
				Bundle extras = intent.getExtras();
				Set<String> keys = extras.keySet();
				{
					// Debug only
					for(String key : keys) {
						Log.d("Deacon", "Got key: " + key +  ", with value = " + extras.get(key));
					}
				}
				if(extras.containsKey("networkInfo")) {
					NetworkInfo netinfo = (NetworkInfo) extras.get("networkInfo");
					if(netinfo.isConnected()) {
						if(!isRunning()){
							Log.d("Deacon","Network has connection!");
							notifyObserversReconnect();
							if(autoRestart) {
								try {
									start(); // <<<<<<< THIS ISN'T WORKING (gets through start() but no pushes received)
								} catch (Exception e) {
									e.printStackTrace();
									Log.d("Deacon", "Deacon tried to start an already-running DeaconService");
								}
							}
						}
					}
					else {
						Log.d("Deacon","Network is disconnected!");
						notifyObserversDisconnect(new DeaconError(new IOException("Lost mobile network connection"), DeaconErrorType.ConnectionError));
						stop();
					}
				}
				else if(extras.containsKey("noConnectivity")) {
					Log.d("Deacon","Network is disconnected!");
					notifyObserversDisconnect(new DeaconError(new IOException("Lost mobile network connection"), DeaconErrorType.ConnectionError));
					// TODO ^^ Copy and paste programming
					stop();
				}
				else {
					Log.d("Deacon", "Received CONNECTIVITY_ACTION intent but no networkInfo or noConnectivty extra data");
				}
		    }
		}
	};
	
	/**
	 * Method to return the broadcast receiver, you must register 
	 * this BroadcastReceiver from the Context.
	 */
	public BroadcastReceiver getBroadcastReceiver(){
		return bcr;
	}
	
	
	/**
	 * @param host The Meteor server to which this client should connect
	 * @param port The client port that the Meteor server is listening on (default is usually 4670) 
	 * @param context The activity that creates this class. Used to register an IntentFilter. // TODO Shouldn't register IntentFilter using context class?
	 * @throws UnknownHostException if host is unreachable
	 * @throws IOException if connection cannot be established
	 * @throws Exception if port value is invalid
	 */
	public Deacon(String host, Integer port, Context context) throws UnknownHostException, IOException, Exception {
		super(host, port);
		this.parent = context;
		this.parent.registerReceiver(this.getBroadcastReceiver(), new IntentFilter("android.net.conn.CONNECTIVITY_CHANGE"));
	}
	
	/**
	 * Called by GC when Deacon object is GC'd. Should be called by application context before Android onDestroy method returns.
	 */
	protected void finalize() {
		Log.d("Deacon", "Shutting down, destroying Deacon object.");
		stop();
		this.parent.unregisterReceiver(bcr);
	}
	
	/**
	 * Android-wrapped version of prase; called whenever a line is received from the Meteor server. 
	 * This version implements Android thread-safe message passing rather than calling parse() directly (from the deaconThread).
	 */
	@Override
	protected void parse(final String meteorMessage) {
		Message msg = Message.obtain(handler, 0, meteorMessage);
		msg.sendToTarget();
	}
	

}
